@@ -7,6 +7,9 @@ const Tp = imports.gi.TelepathyGLib;
 
 const ChatroomManager = imports.chatroomManager;
 const Lang = imports.lang;
+const Mainloop = imports.mainloop;
+
+const MAX_USERS_SHOWN = 8;
 
 const UserListPopover = new Lang.Class({
     Name: 'UserListPopover',
@@ -19,30 +22,21 @@ const UserListPopover = new Lang.Class({
         }));
         this.widget.connect('map', Lang.bind(this, function() {
             this._revealer.transition_duration = 0;
+            this._ensureUserList();
         }));
         this._revealer.connect('notify::child-revealed', Lang.bind(this, function() {
             this._revealer.transition_duration = 250;
         }));
 
-        this._rooms = {};
-        this._room = null;
-
         this._roomManager = new ChatroomManager.getDefault();
-        this._roomManager.connect('room-added',
-                                  Lang.bind(this, this._roomAdded));
-        this._roomManager.connect('room-removed',
-                                  Lang.bind(this, this._roomRemoved));
         this._roomManager.connect('active-changed',
                                   Lang.bind(this, this._activeRoomChanged));
     },
 
     _createWidget: function() {
         this.widget = new Gtk.Popover({ modal: true,
-                                        position: Gtk.PositionType.TOP,
-                                        vexpand: true,
-                                        margin_start: 12,
-                                        margin_end: 12,
-                                        margin_bottom: 12 });
+                                        position: Gtk.PositionType.TOP });
+
         this.widget.set_border_width(6);
         this.widget.set_size_request(250, -1);
 
@@ -72,55 +66,43 @@ const UserListPopover = new Lang.Class({
         this._box.show_all();
     },
 
-    _roomAdded: function(roomManager, room) {
-        if (room.type != Tp.HandleType.ROOM)
-            return;
-
-        let userList = new UserList(room);
-        this._rooms[room.id] = userList;
-
-        userList.widget.vadjustment.connect('changed',
-                                            Lang.bind(this, this._updateEntryVisibility));
-    },
-
-    _roomRemoved: function(roomManager, room) {
-        if (!this._rooms[room.id])
-            return;
-        this._rooms[room.id].widget.destroy();
-        delete this._rooms[room.id];
-    },
-
     _activeRoomChanged: function(manager, room) {
         this._entry.text = '';
-        this._updateFilter();
 
-        let currentList = this._room ? this._rooms[this._room.id] : null;
-        if (currentList)
-            this._box.remove(currentList.widget);
+        if (this._userList)
+            this._userList.widget.destroy();
+        this._userList = null;
+    },
 
-        this._room = room;
-
-        if (!room || !this._rooms[room.id])
+    _ensureUserList: function() {
+        if (this._userList)
             return;
 
-        this._box.add(this._rooms[room.id].widget);
+        let room = this._roomManager.getActiveRoom();
+        if (!room || room.type != Tp.HandleType.ROOM)
+            return;
+
+        this._userList = new UserList(room);
+        this._box.add(this._userList.widget);
+
+        this._userList.widget.vadjustment.connect('changed',
+                                                  Lang.bind(this, this._updateEntryVisibility));
         this._updateEntryVisibility();
     },
 
     _updateEntryVisibility: function() {
-        if (!this._room || !this._rooms[this._room.id])
+        if (!this._userList)
             return;
-        let userList = this._rooms[this._room.id];
-        let [, natHeight] = userList.widget.get_child().get_preferred_height();
-        let height = userList.widget.get_allocated_height();
-        this._revealer.reveal_child = this._entry.text != '' ||
-                                      natHeight > height;
+
+        let reveal = this._entry.text != '' ||
+                     this._userList.numRows > MAX_USERS_SHOWN;
+        this._revealer.reveal_child = reveal;
     },
 
     _updateFilter: function() {
-        if (!this._room || !this._rooms[this._room.id])
+        if (!this._userList)
             return;
-        this._rooms[this._room.id].setFilter(this._entry.text);
+        this._userList.setFilter(this._entry.text);
     }
 });
 
@@ -147,6 +129,7 @@ const UserListRow = new Lang.Class({
     },
 
     set expand(expand) {
+        this._ensureDetails();
         this._revealer.reveal_child = expand;
     },
 
@@ -170,8 +153,14 @@ const UserListRow = new Lang.Class({
         this._revealer = new Gtk.Revealer({ reveal_child: false });
         vbox.add(this._revealer);
 
+        this.widget.show_all();
+    },
+
+    _ensureDetails: function() {
+        if (this._revealer.get_child())
+            return;
+
         let frame = new Gtk.Frame({ hexpand: true });
-        this._revealer.add(frame);
 
         let box = new Gtk.Box({ orientation: Gtk.Orientation.VERTICAL,
                                 spacing: 6, margin: 6 });
@@ -189,6 +178,7 @@ const UserListRow = new Lang.Class({
                                            hexpand: true });
         box.add(this._detailsGrid);
 
+        let user = this.widget.user;
         if (user != user.connection.self_contact) {
             let button = new Gtk.Button({ label: _("Message"),
                                           margin_top: 12,
@@ -202,7 +192,8 @@ const UserListRow = new Lang.Class({
             box.add(button);
         }
 
-        this.widget.show_all();
+        this._revealer.add(frame);
+        frame.show_all();
     },
 
     _formatLast: function(seconds) {
@@ -248,6 +239,8 @@ const UserListRow = new Lang.Class({
         if (!fn)
             fn = this.widget.user.alias;
 
+        this._detailsGrid.foreach(function(w) { w.destroy(); });
+
         let row = 0;
         let w = new Gtk.Label({ label: fn, ellipsize: Pango.EllipsizeMode.END,
                                 halign: Gtk.Align.START });
@@ -297,11 +290,10 @@ const UserListRow = new Lang.Class({
             this.widget.get_style_context().add_class('expanded');
             this._arrow.arrow_type = Gtk.ArrowType.DOWN;
 
-            this._spinnerBox.show();
+            let prevDetails = this._detailsGrid.get_children().length > 0;
+            this._spinnerBox.visible = !prevDetails;
             this._spinner.start();
 
-            this._detailsGrid.foreach(function(w) { w.destroy(); });
-
             this._cancellable = new Gio.Cancellable();
             this.widget.user.request_contact_info_async(this._cancellable,
                                                         Lang.bind(this, this._onContactInfoReady));
@@ -323,13 +315,17 @@ const UserList = new Lang.Class({
     Name: 'UserList',
 
     _init: function(room) {
-        this.widget = new Gtk.ScrolledWindow({ hexpand: true, vexpand: true,
+        this.widget = new Gtk.ScrolledWindow({ hexpand: true,
                                                shadow_type: Gtk.ShadowType.ETCHED_IN });
         this.widget.hscrollbar_policy = Gtk.PolicyType.NEVER;
 
-        this._list = new Gtk.ListBox();
+        this._list = new Gtk.ListBox({ vexpand: true });
         this.widget.add(this._list);
 
+        this._updateHeightId = 0;
+        this._list.connect('size-allocate',
+                           Lang.bind(this, this._updateContentHeight));
+
         this._list.set_selection_mode(Gtk.SelectionMode.NONE);
         /* see https://bugzilla.gnome.org/show_bug.cgi?id=725403 */
         //this._list.set_header_func(Lang.bind(this, this._updateHeader));
@@ -338,39 +334,77 @@ const UserList = new Lang.Class({
 
         this._list.connect('row-activated',
                            Lang.bind(this, this._onRowActivated));
+        this.widget.connect('destroy',
+                            Lang.bind(this, this._onDestroy));
 
         this._room = room;
         this._rows = {};
         this._activeRow = null;
 
-        room.connect('member-renamed',
-                     Lang.bind(this, this._onMemberRenamed));
-        room.connect('member-disconnected',
-                     Lang.bind(this, this._onMemberRemoved));
-        room.connect('member-kicked',
-                     Lang.bind(this, this._onMemberRemoved));
-        room.connect('member-banned',
-                     Lang.bind(this, this._onMemberRemoved));
-        room.connect('member-left',
-                     Lang.bind(this, this._onMemberRemoved));
-        room.connect('member-joined',
-                     Lang.bind(this, this._onMemberJoined));
-        /*
-        // see https://bugzilla.gnome.org/show_bug.cgi?id=725403
-        room.connect('members-changed',
-                     Lang.bind(this, this._onMembersChanged));
-        */
-        room.connect('notify::channel',
-                     Lang.bind(this, this._onChannelChanged));
+        let roomSignals = [
+            { name: 'member-renamed',
+              handler: Lang.bind(this, this._onMemberRenamed) },
+            { name: 'member-disconnected',
+              handler: Lang.bind(this, this._onMemberRemoved) },
+            { name: 'member-kicked',
+              handler: Lang.bind(this, this._onMemberRemoved) },
+            { name: 'member-banned',
+              handler: Lang.bind(this, this._onMemberRemoved) },
+            { name: 'member-left',
+              handler: Lang.bind(this, this._onMemberRemoved) },
+            { name: 'member-joined',
+              handler: Lang.bind(this, this._onMemberJoined) },
+            /*
+            // see https://bugzilla.gnome.org/show_bug.cgi?id=725403
+            { name: 'members-changed',
+              handler: Lang.bind(this, this._onMembersChanged) },
+            */
+            { name: 'notify::channel',
+              handler: Lang.bind(this, this._onChannelChanged) }
+        ];
+        this._roomSignals = [];
+        roomSignals.forEach(Lang.bind(this, function(signal) {
+            this._roomSignals.push(room.connect(signal.name, signal.handler));
+        }));
+        this._onChannelChanged(room);
 
         this.widget.show_all();
     },
 
+    get numRows() {
+        return Object.keys(this._rows).length;
+    },
+
+    _onDestroy: function() {
+        for (let i = 0; i < this._roomSignals.length; i++)
+            this._room.disconnect(this._roomSignals[i]);
+        this._roomSignals = [];
+    },
+
     setFilter: function(filter) {
         this._filter = filter.toLowerCase();
         this._list.invalidate_filter();
     },
 
+    _updateContentHeight: function() {
+        if (this._updateHeightId != 0)
+            return;
+
+        this._updateHeightId = Mainloop.idle_add(Lang.bind(this, function() {
+            let topRow = this._list.get_row_at_y(this.widget.vadjustment.value);
+            let membersShown = Math.min(this.numRows, MAX_USERS_SHOWN);
+            let index = Math.min(topRow.get_index(), this.numRows - membersShown);
+            let height = 0;
+
+            for (let i = 0; i < membersShown; i++)
+                height += this._list.get_row_at_index(index + i).get_allocated_height();
+
+            this.widget.min_content_height = height;
+            this._updateHeightId = 0;
+            return GLib.SOURCE_REMOVE;
+        }));
+    },
+
     _onMemberRenamed: function(room, oldMember, newMember) {
         this._removeMember(oldMember);
         this._addMember(newMember);
@@ -385,8 +419,7 @@ const UserList = new Lang.Class({
     },
 
     _onMembersChanged: function(room) {
-        let numMembers = room.channel.group_dup_members_contacts().length;
-        this._counterLabel.label = numMembers.toString();
+        this._counterLabel.label = this.numRows.toString();
     },
 
     _onChannelChanged: function(room) {
@@ -409,8 +442,8 @@ const UserList = new Lang.Class({
 
     _removeMember: function(member) {
         let row = this._rows[member];
-        if (row && row.widget.get_parent())
-            this._list.remove(row.widget);
+        if (row)
+            row.widget.destroy();
         delete this._rows[member];
     },
 
@@ -447,9 +480,6 @@ const UserList = new Lang.Class({
         if (!this._room.channel)
             return;
 
-        let members = this._room.channel.group_dup_members_contacts();
-        let numMembers = members.length;
-
         let box = new Gtk.Box({ orientation: Gtk.Orientation.HORIZONTAL,
                                 margin_start: 6,
                                 margin_end: 6,
@@ -458,7 +488,7 @@ const UserList = new Lang.Class({
                                 use_markup: true,
                                 hexpand: true,
                                 halign: Gtk.Align.START }));
-        this._counterLabel = new Gtk.Label({ label: numMembers.toString(),
+        this._counterLabel = new Gtk.Label({ label: this.numRows.toString(),
                                              halign: Gtk.Align.END });
         box.add(this._counterLabel);
         box.show_all();
